/* nttest.c: -*- C -*-  DESCRIPTIVE TEXT. */

/*  Copyright (c) 1996 Brian J. Fox
    Author: Brian J. Fox (bfox@ai.mit.edu) Mon Jan  6 22:29:30 1997.  */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <sys/types.h>
#include <sys/fcntl.h>
#include <sys/stat.h>

int
main (int argc, char *argv[])
{
  char *buffer;
  struct stat finfo;
  int fd;
  char *file = "/www/winbone/docs/System/images/site/Contact.gif";

  stat (file, &finfo);
  buffer = (char *)malloc (finfo.st_size);
  setmode (fileno (stdout), O_BINARY);
  fd = open (file, O_RDONLY | O_BINARY, 0666);
  read (fd, buffer, finfo.st_size);
  fprintf (stdout, "Content-Type: image/gif\n\n");
  write (fileno (stdout), buffer, finfo.st_size);
  return (0);
}

  

  
